//
// Created by Ari Trachtenberg on 9/10/24.
//

#ifndef PUBLIC_COMMON_H
#define PUBLIC_COMMON_H

#include <iostream>
#include <string>
#include <sstream>
#include <regex>
#include <cmath>
#include <limits>

using namespace std;

/**
 * A Singleton class of variables and functions that are common to multiple labs.
 */

class Common {
//
// Created by Ari Trachtenberg on 9/3/24, with support from ChatGPT
//

public:

// global variables
    static streambuf *orig_cin_buf;  // records the standard input buffer
    static streambuf *orig_cout_buf; // records the standard output buffer
    static istringstream inStream;   // input stream to produce from a given string
    static ostringstream outStream;  // captured output stream


// HELPERS

// Function to trim leading and trailing whitespace - from ChatGPT
    static string trim(const string &str) {
        size_t first = str.find_first_not_of(" \t\n\r\f\v");
        if (first == std::string::npos)
            return "";  // The string is all whitespace or empty

        size_t last = str.find_last_not_of(" \t\n\r\f\v");
        return str.substr(first, (last - first + 1));
    }

    /**
     * Attempts to extract the last viable floating point number from a string.
     * @param str The string to search.
     * @return A floating point number that was found
     * Based on ChatGPT output
     */
    static float getFloat(string str) {
        std::regex floatRe(R"(-?\d+(\.\d+)?)");
        std::sregex_iterator currentMatch(str.begin(), str.end(), floatRe);
        std::sregex_iterator lastMatch; // end iterator

        std::string lastFloat;

        while (currentMatch != lastMatch) {
            lastFloat = currentMatch->str();  // Update with the current match
            ++currentMatch;  // Move to the next match
        }

        return stof(lastFloat);  // Return the last matched float
    }

    /**
  * Attempts to extract the last viable int from a string.
  * @param str The string to search.
  * @return A int that was found
  * Based on ChatGPT output
  */
    static int getInt(string str) {
        std::regex intRe(R"(-?\d+)");
        std::sregex_iterator currentMatch(str.begin(), str.end(), intRe);
        std::sregex_iterator lastMatch; // end iterator

        std::string lastInt;

        while (currentMatch != lastMatch) {
            lastInt = currentMatch->str();  // Update with the current match
            ++currentMatch;  // Move to the next match
        }

        try {
            return stoi(lastInt);  // Return the last matched float
        } catch (exception& ignored) {
            cerr << "I could not parse " << lastInt << " into an integer." << endl;
            return numeric_limits<int>::min();
        }
    }

/**
 * Redirects input and output to strings
 * @param iStr The input string
 */
    static void redirIO(const string iStr) {
        // save the old buffers
        orig_cin_buf = cin.rdbuf();
        orig_cout_buf = cout.rdbuf();

        // set up streams
        inStream.str(iStr);
        outStream.str("");
        outStream.clear();
        cin.rdbuf(inStream.rdbuf());
        cout.rdbuf(outStream.rdbuf());
    }

/**
 * Restore input and output to STDIN/STDOUT, respectively.
 * @requires redirIO() was called before this method
 * @return The output string that was captured since the redirect, trimmed of
 *         initial and final whitespace
 */
    static string restoreIO() {
        cin.rdbuf(orig_cin_buf);
        cout.rdbuf(orig_cout_buf);

        return trim(outStream.str());
    }

    // Testing functions
    /**
   * Test a given function and returns a float result
   * @param func A reference to the function. to be evaluated
   * @param input The input to be provided, through stdin, to the function.
   * @return The contents of what func directs to standard output
   */
    static string testFunc(void (*func)(void), string input) {
        redirIO(input);
        func();
        return restoreIO();
    }

/**
 * Same as testFuncFloat, but with a float output
 */
    static float testFuncFloat(void (*func)(void), string input) {
        return getFloat(testFunc(func, input));
    }

/**
 * Same as testFuncFloat, but with an int output
 */
    static int testFuncInt(void (*func)(void), string input) {
        return getInt(testFunc(func, input));
    }
};

#endif //PUBLIC_COMMON_H
